/*
	Ritma - Advanced metronome for GBA
	Copyright (C) 2005  Avelino Herrera Morales

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "gba.h"
#include "font_8x8.h"

#define  PRESSED(k)  (!(REG_KEYPAD & (k)))

#define  RANDOM_0_100  (rand() % 101)

int __gba_multiboot;

/* modo grfico */
#define BG0_ENABLE      0x100
#define BG1_ENABLE      0x200
#define BG2_ENABLE      0x400
#define BG3_ENABLE      0x800
#define OBJ_ENABLE      0x1000
#define WIN0_ENABLE     0x2000
#define WIN1_ENABLE     0x4000
#define WINOBJ_ENABLE   0x8000

#define BACK_BUFFER     0x10

#define MODE0           0x0
#define MODE1           0x1
#define MODE2           0x2
#define MODE3           0x3
#define MODE4           0x4
#define MODE5           0x5

#define  VIDEO_BUFFER_0  ((u16 *) 0x6000000)
#define  VIDEO_BUFFER_1  ((u16 *) 0x600A000)
#define  BG_PALETTE_MEM  ((u16 *) 0x5000000)

#define  TOP_LEFT      0
#define  TOP_RIGHT     239
#define  BOTTOM_LEFT   38060
#define  BOTTOM_RIGHT  38299

#define RGB16(r,g,b)  ((r) + ((g) << 5) + ((b) << 10))

#define  ERASE_SCREEN(scr)  memset(scr, 0, 240 * 160)

#define  MAX_BPM  300
#define  MIN_BPM  10

#define  MAX_BEAT  16
#define  MIN_BEAT  1

u16 *current_video_buffer = VIDEO_BUFFER_0;
u16 current_bpm = 120;
u16 sound_enable = 0;
u16 beat_size = 4;
u16 current_beat = 0;
u16 decade_enable = 0;
u16 timer_configured = 0;

#define  GET_N(f)     (2048 - (131072 / (f)))
#define  PLAY_TICK_A  {REG_SOUND1CNT_L = 0x0000; REG_SOUND1CNT_H = 60 | SOUNDDUTY50 | SOUND1ENVINIT(15); REG_SOUND1CNT_X = GET_N(2000) | SOUND1PLAYONCE | SOUNDINIT;}
#define  PLAY_TICK_B  {REG_SOUND1CNT_L = 0x0000; REG_SOUND1CNT_H = 60 | SOUNDDUTY50 | SOUND1ENVINIT(15); REG_SOUND1CNT_X = GET_N(1000) | SOUND1PLAYONCE | SOUNDINIT;}


void InterruptProcess(void) __attribute__ ((section(".iwram")));   // the ISR must be declared in this way

 
void InterruptProcess(void) {
	if (sound_enable) {
		if (current_beat == 0) {
			PLAY_TICK_A;
			current_beat++;
			if (current_beat >= beat_size)
				current_beat = 0;
		}
		else {
			PLAY_TICK_B;
			current_beat++;
			if (current_beat >= beat_size)
				current_beat = 0;
		}
	}
	if (!timer_configured) {
		// reload timer 0
		REG_TM0CNT_H = 0;
		REG_TM0CNT_L = 0xFFFF - (u16) (((u32)60 * (u32)16387) / (u32)current_bpm);  // set BPM
		REG_TM0CNT_H = 0x80 | 0x40 | 0x03;                                 // start + irq + f/1024
		timer_configured = 1;
	}
	// clear interrupt flag
	REG_IF |= REG_IF;
}


void flip_and_show(void) {
	if (current_video_buffer == VIDEO_BUFFER_1) {
		REG_DISPCNT |= BACK_BUFFER;
		current_video_buffer = VIDEO_BUFFER_0;
	}
	else {
		REG_DISPCNT &= ~BACK_BUFFER;
		current_video_buffer = VIDEO_BUFFER_1;
	}
}


void put_character(char c, int x, int y, u16 *video_buffer, int erase) {
	int font_y = (((int) c) - 32) / 16;
	int font_x = (((int) c) - 32) % 16;
	int i, j;
	int font_offset = (font_y * 64 * 8) + (font_x * 4);
	int offset = ((y * 240) + x) >> 1;
	for (i = 0; i < 8; i++, offset += 120, font_offset += 64) {
		if (erase) {
			for (j = 0; j < 4; j++)
				video_buffer[offset + j] = 0x0000;
		}
		for (j = 0; j < 4; j++) {
			register u16 d = fontData[font_offset + j];
			register u16 prev = video_buffer[offset + j];
			register u16 to_write = 0;
			if ((d & 0x00FF) != fontTransparentIndex)
				to_write |= (d & 0x00FF);
			else
				to_write |= (prev & 0x00FF);
			if ((d >> 8) != fontTransparentIndex)
				to_write |= (d & 0xFF00);
			else
				to_write |= (prev & 0xFF00);
			video_buffer[offset + j] = to_write;
		}
	}
}


void put_string(char *s, int x, int y, u16 *video_buffer, int erase) {
	int i = 0;
	while (s[i] != 0) {
		put_character(s[i], x, y, video_buffer, erase);
		x += 8;
		i++;
	}
}


void wait_for_vsync() {
	while (((volatile u16) REG_VCOUNT) != 160)
		;
}


#define  GET_X_CENTER(width)  (120 - ((width) * 4))

void show_initial_screen() {
	u32 random_seed = 0;

	ERASE_SCREEN(current_video_buffer);
	put_string("R I T M A", GET_X_CENTER(9), 30, current_video_buffer, 0);
	put_string("@ 2006", GET_X_CENTER(6), 50, current_video_buffer, 0);
	put_string("Avelino Herrera Morales", GET_X_CENTER(23), 59, current_video_buffer, 0);
	put_string("http://gba.gabiot.com", GET_X_CENTER(21), 89, current_video_buffer, 0);
	put_string("PRESS START", GET_X_CENTER(11), 110, current_video_buffer, 0);
	wait_for_vsync();
	flip_and_show();
	while (!PRESSED(BUTTON_START))
		random_seed++;
	srand(random_seed);
	while (PRESSED(BUTTON_START))
		;
	ERASE_SCREEN(current_video_buffer);
	wait_for_vsync();
	flip_and_show();
}


void repaint_current_bpm() {
	char buffer[20];
	sprintf(buffer, "%03d bpm   beat=%02d", current_bpm, beat_size);
	put_string(buffer, GET_X_CENTER(17), 40, current_video_buffer, 0);
}


void show_main_screen() {
	ERASE_SCREEN(current_video_buffer);
	repaint_current_bpm();
	put_string("RITMA v0.3", GET_X_CENTER(10), 10, current_video_buffer, 0);
	put_string("   Up/Down: tempo     ", GET_X_CENTER(22), 60, current_video_buffer, 0);
	put_string("Left/Right: beat      ", GET_X_CENTER(22), 70, current_video_buffer, 0);
	put_string("         A: decades   ", GET_X_CENTER(22), 80, current_video_buffer, 0);
	put_string("     Start: start/stop", GET_X_CENTER(22), 90, current_video_buffer, 0);
	if (sound_enable)
		put_string("PLAYING", GET_X_CENTER(7), 120, current_video_buffer, 0);
	else
		put_string("STOPED ", GET_X_CENTER(7), 120, current_video_buffer, 0);
	wait_for_vsync();
	flip_and_show();
}


void play_tick_a() {
	REG_SOUND1CNT_L = 0x0000;
	REG_SOUND1CNT_H = 60 | SOUNDDUTY50 | SOUND1ENVINIT(15);
	REG_SOUND1CNT_X = GET_N(2000) | SOUND1PLAYONCE | SOUNDINIT;
	/* channel 4
	REG_SOUND4CNT_L = 60 | SOUND4PLAYONCE;
	REG_SOUND4CNT_H = 0x8033 | 0x4000;
	*/
}


void play_tick_b() {
	REG_SOUND1CNT_L = 0x0000;
	REG_SOUND1CNT_H = 60 | SOUNDDUTY50 | SOUND1ENVINIT(15);
	REG_SOUND1CNT_X = GET_N(1000) | SOUND1PLAYONCE | SOUNDINIT;
	/* channel 4
	REG_SOUND4CNT_L = 63 | SOUND4PLAYONCE;
	REG_SOUND4CNT_H = 0x8032 | 0x4000;
	*/
}


void stop_beat() {
	sound_enable = 0;
	timer_configured = 0;
	REG_IME = 0;
	REG_TM0CNT_H = 0;
}


void start_beat() {
	if (sound_enable)
		return;
	current_beat = 0;
	sound_enable = 1;
	// stop timer 0
	REG_TM0CNT_H = 0;
	timer_configured = 0;
	// enable interrupt for timer 0
	REG_IE = 0x0008;
	REG_IME = 1;
	// restart timer 0
	//REG_TM0CNT_L = 0xFFFF - (u16) (((u32)60 * (u32)16387) / (u32)current_bpm);  // set BPM
	REG_TM0CNT_L = 0xFFFF;
	REG_TM0CNT_H = 0x80 | 0x40 | 0x03;                        // start + irq + f/1024
}


int AgbMain(void) {
	int i;

	REG_DISPCNT = MODE4 | BG2_ENABLE;
	for (i = 0; i < 256; i++)
		BG_PALETTE_MEM[i] = fontPalette[i];

	show_initial_screen();

	// enable audio output for sound channed 1
	REG_SOUNDCNT_X = 0x80;
	REG_SOUNDCNT_L = 0x1177;
	REG_SOUNDCNT_H = 2;

	while (1) {
		if (PRESSED(BUTTON_START)) {
			if (sound_enable)
				stop_beat();
			else
				start_beat();
			while (PRESSED(BUTTON_START))
				;
		}
		if (PRESSED(BUTTON_A))
			decade_enable = 1;
		else
			decade_enable = 0;
		if (PRESSED(KEYPAD_UP)) {
			s16 temp = 0;
			if (decade_enable)
				temp = current_bpm + 10;
			else
				temp = current_bpm + 1;
			if (temp > MAX_BPM)
				temp = MAX_BPM;
			current_bpm = (u16) temp;
			timer_configured = 0;
			while (PRESSED(KEYPAD_UP))
				;
		}
		else if (PRESSED(KEYPAD_DOWN)) {
			s16 temp = 0;
			if (decade_enable)
				temp = current_bpm - 10;
			else
				temp = current_bpm - 1;
			if (temp < MIN_BPM)
				temp = MIN_BPM;
			current_bpm = (u16) temp;
			timer_configured = 0;
			while (PRESSED(KEYPAD_DOWN))
				;
		}
		if (PRESSED(KEYPAD_LEFT)) {
			s16 temp = beat_size;
			temp--;
			if (temp < MIN_BEAT)
				temp = MIN_BEAT;
			beat_size = (u16) temp;
			while (PRESSED(KEYPAD_LEFT))
				;
		}
		else if (PRESSED(KEYPAD_RIGHT)) {
			s16 temp = beat_size;
			temp++;
			if (temp > MAX_BEAT)
				temp = MAX_BEAT;
			beat_size = (u16) temp;
			while (PRESSED(KEYPAD_RIGHT))
				;
		}
		show_main_screen();
	}
}

